;;; usart driver

;; bit definitions

; UCSRA
.equ RXC = 7
.equ TXC = 6
.equ UDRE = 5
.equ FE = 4
.equ DOR = 3
.equ PE = 2
.equ U2X = 1
.equ MPCM = 0

; UCSRB
.equ RXCIE = 7
.equ TXCIE = 6
.equ UDRIE = 5
.equ RXEN = 4
.equ TXEN = 3
.equ UCSZ2 = 2
.equ RXB8 = 1
.equ TXB8 = 0

; UCSRC
.equ UMSEL = 6
.equ UPM1 = 5
.equ UPM0 = 4
.equ USBS = 3
.equ UCSZ1 = 2
.equ UCSZ0 = 1
.equ UCPOL = 0


.macro usart_driver

;; data buffer

.equ usart@0_tx_mask = usart@0_tx_size - 1
.equ usart@0_rx_mask = usart@0_rx_size - 1

.set usart@0_tx_in = heap
.set heap = heap + 1

.set usart@0_tx_out = heap
.set heap = heap + 1

.set usart@0_tx_data = heap
.set heap = heap + usart@0_tx_size

.set usart@0_rx_in = heap
.set heap = heap + 1

.set usart@0_rx_out = heap
.set heap = heap + 1

.set usart@0_rx_data = heap
.set heap = heap + usart@0_rx_size


;; usart@0
;;
;; ( -- )
;;
;; enable usart

link_
.db "usart@0",0,0

global usart@0:
  out_ usart@0_tx_in,zero
  out_ usart@0_tx_out,zero
  out_ usart@0_rx_in,zero
  out_ usart@0_rx_out,zero
  ldi temp0,(1<<TXEN)|(1<<RXEN)|(1<<RXCIE)
  out_ UCSR@0B,temp0
  ret


;; /usart@0
;;
;; ( -- )
;;
;; disable usart

link_
.db "/usart@0",0

global notusart@0:
  out_ UCSR@0B,zero
  ret


;; baud@0
;;
;; ( u -- )
;;
;; set baud divisor

link_
.db "baud@0",0

global baud@0:
  out_ UBRR@0L,tosl
  out_ UBRR@0H,tosh
  drop_
  ret


;; tx@0
;;
;; ( c -- )
;;
;; send character to usart

link_
.db "tx@0",0

global tx@0:
  in_ xl,usart@0_tx_in
  inc xl
  andi xl,usart@0_tx_mask

  in_ xh,usart@0_tx_out
  cp xh,xl
  brne tx@0_store

  call pause
  rjmp tx@0
  
tx@0_store:
  out_ usart@0_tx_in,xl
  ldi zl,low(usart@0_tx_data)
  ldi zh,high(usart@0_tx_data)
  add zl,xl
  adc zh,zero
  st z,tosl
  drop_
  sbi_ UCSR@0B,UDRIE,temp0
  ret


;; tx@0?
;;
;; ( -- ; f )
;;
;; test if character can be sent

link_
.db "tx@0?",0,0

global tx@0q:
  in_ xl,usart@0_tx_in
  inc xl
  andi xl,usart@0_tx_mask
  in_ xh,usart@0_tx_out
  cp xh,xl
  ret


;; udre isr

.set pc_ = pc
.org USART@0_UDRE
  rjmp usart@0_udre_isr
.org pc_
  
usart@0_udre_isr:
  push xl
  in_ xl,SREG
  push xl
  push xh
  push zl
  push zh

  in_ xl,usart@0_tx_in
  in_ xh,usart@0_tx_out

  cp xh,xl
  brne usart@0_udre_next

usart@0_udre_last:
  cbi_ UCSR@0B,UDRIE,xl
  rjmp usart@0_udre_done

usart@0_udre_next:
  inc xh
  andi xh,usart@0_tx_mask
  out_ usart@0_tx_out,xh
  
  ldi zl,low(usart@0_tx_data)
  ldi zh,high(usart@0_tx_data)
  add zl,xh
  adc zh,zero
  
  ld xl,z
  out_ UDR@0,xl

usart@0_udre_done:
  pop zh
  pop zl
  pop xh
  pop xl
  out_ SREG,xl
  pop xl
  reti


;; rx@0
;;
;; ( -- c )
;;
;; fetch character from uart

link_
.db "rx@0",0

global rx@0:
  in_ xh,usart@0_rx_out

  in_ xl,usart@0_rx_in
  cp xh,xl
  brne rx@0_fetch

  call pause
  rjmp rx@0
  
rx@0_fetch:
  inc xh
  andi xh,usart@0_rx_mask
  out_ usart@0_rx_out,xh

  ldi zl,low(usart@0_rx_data)
  ldi zh,high(usart@0_rx_data)
  add zl,xh
  adc zh,zero

  dup_
  ld tosl,z
  clr tosh
  ret


;; rx@0?
;;
;; ( -- ; f )
;;
;; test if character is available

link_
.db "rx@0?",0,0

global rx@0q:
  in_ xh,usart@0_rx_out
  in_ xl,usart@0_rx_in
  cp xh,xl
  ret

  
;; rxc isr

.set pc_ = pc
.org USART@0_RXC
  rjmp usart@0_rxc_isr
.org pc_

usart@0_rxc_isr:
  push xl
  in_ xl,SREG
  push xl
  push xh
  push zl
  push zh

  in_ xl,usart@0_rx_in
  in_ xh,usart@0_rx_out

  inc xl
  andi xl,usart@0_rx_mask
  cp xh,xl
  brne usart@0_rxc_next

usart@0_rxc_full:
  in_ xl,UDR@0
  rjmp usart@0_rxc_done

usart@0_rxc_next:
  out_ usart@0_rx_in,xl

  ldi zl,low(usart@0_rx_data)
  ldi zh,high(usart@0_rx_data)
  add zl,xl
  adc zh,zero

  in_ xl,UDR@0
  st z,xl

usart@0_rxc_done:
  pop zh
  pop zl
  pop xh
  pop xl
  out_ SREG,xl
  pop xl
  reti

.endmacro


.ifdef UDR0
  usart_driver 0
.endif

.ifdef UDR1
  usart_driver 1
.endif

